import zope.interface
import zope.schema


class IBookmark(zope.interface.Interface):
    cmpl = zope.schema.Complex(required=False)
    real = zope.schema.Real(required=False)
    integer = zope.schema.Int(required=False)


@zope.interface.implementer(IBookmark)
class Bookmark(object):
    cmpl = None
    real = None
    integer = None


def main() -> None:
    bm: IBookmark = Bookmark()
    bm.cmpl = "str"  # Error: number is expected
    bm.cmpl = complex(1, 2)
    bm.cmpl = 1.2
    bm.cmpl = 343
    bm.real = complex(1, 2)  # Error: real is expected
    bm.real = 1.2
    bm.real = 343
    bm.integer = complex(1, 2)  # Error: integer is expected
    bm.integer = 1.2  # Error: integer is expected
    bm.integer = 343


if __name__ == '__main__':
    main()

"""
<output>
schema_number.py:20: error: Incompatible types in assignment (expression has type "str", variable has type "Optional[complex]")
schema_number.py:24: error: Incompatible types in assignment (expression has type "complex", variable has type "Optional[float]")
schema_number.py:27: error: Incompatible types in assignment (expression has type "complex", variable has type "Optional[int]")
schema_number.py:28: error: Incompatible types in assignment (expression has type "float", variable has type "Optional[int]")
</output>
"""



# Stubs for zope.interface.common.interfaces (Python 3.6)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from zope.interface import Interface

class IException(Interface): ...
class IStandardError(IException): ...
class IWarning(IException): ...
class ISyntaxError(IStandardError): ...
class ILookupError(IStandardError): ...
class IValueError(IStandardError): ...
class IRuntimeError(IStandardError): ...
class IArithmeticError(IStandardError): ...
class IAssertionError(IStandardError): ...
class IAttributeError(IStandardError): ...
class IDeprecationWarning(IWarning): ...
class IEOFError(IStandardError): ...
class IEnvironmentError(IStandardError): ...
class IFloatingPointError(IArithmeticError): ...
class IIOError(IEnvironmentError): ...
class IImportError(IStandardError): ...
class IIndentationError(ISyntaxError): ...
class IIndexError(ILookupError): ...
class IKeyError(ILookupError): ...
class IKeyboardInterrupt(IStandardError): ...
class IMemoryError(IStandardError): ...
class INameError(IStandardError): ...
class INotImplementedError(IRuntimeError): ...
class IOSError(IEnvironmentError): ...
class IOverflowError(IArithmeticError): ...
class IOverflowWarning(IWarning): ...
class IReferenceError(IStandardError): ...
class IRuntimeWarning(IWarning): ...
class IStopIteration(IException): ...
class ISyntaxWarning(IWarning): ...
class ISystemError(IStandardError): ...
class ISystemExit(IException): ...
class ITabError(IIndentationError): ...
class ITypeError(IStandardError): ...
class IUnboundLocalError(INameError): ...
class IUnicodeError(IValueError): ...
class IUserWarning(IWarning): ...
class IZeroDivisionError(IArithmeticError): ...



# Stubs for zope.schema._compat (Python 3.6)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any

PY3: Any
PY2: Any
string_types: Any
text_type = str
binary_type = bytes
integer_types: Any

def non_native_string(x: Any): ...
def make_binary(x: Any): ...
text_type = unicode
binary_type = str



"""zope.interface provides an abstract attribute for classes

It is translated to Any type.
"""
import zope.interface

class IBookmark(zope.interface.Interface):
    field = zope.interface.Attribute('Arbitrary Attribute')


@zope.interface.implementer(IBookmark)
class Bookmark(object):
    field = None


def main() -> None:
    bm: IBookmark = Bookmark()

    # We can assign anything to abstract attributes
    bm.field = 343
    bm.field = None
    bm.field = "Sample"

if __name__ == '__main__':
    main()

"""
<output>
</output>
"""



"""zope.interface provides an abstract attribute for classes

It is translated to Any type.
"""
import zope.interface
from typing import Optional

class IBookmark(zope.interface.Interface):
    field = zope.interface.Attribute('Arbitrary Attribute')


@zope.interface.implementer(IBookmark)
class Bookmark(object):
    field = None

def main(obj: Optional[IBookmark]) -> None:

    if not IBookmark.providedBy(obj):
        reveal_type(obj)
    else:
        reveal_type(obj)
    reveal_type(obj)

    cls = obj.__class__
    if IBookmark.implementedBy(cls):
        reveal_type(cls)
    else:
        reveal_type(cls)
    reveal_type(cls)

if __name__ == '__main__':
    main(Bookmark())

"""
<output>
interface_implications.py:19: note: Revealed type is 'None'
interface_implications.py:21: note: Revealed type is '__main__.IBookmark'
interface_implications.py:22: note: Revealed type is 'Union[__main__.IBookmark, None]'
interface_implications.py:26: note: Revealed type is 'Type[__main__.IBookmark*]'
interface_implications.py:28: note: Revealed type is 'Union[Type[__main__.IBookmark*], Type[None]]'
interface_implications.py:29: note: Revealed type is 'Union[Type[__main__.IBookmark*], Type[None]]'
 </output>
"""



"""Interfaces can be inherited
"""
import zope.interface


class ISomething(zope.interface.Interface):
    def hello(x: int, y: str) -> None:
        pass

class ISuperSomething(ISomething):
    def greet(msg: str) -> None:
        pass

class ISomethingSomething(ISuperSomething):
    def wink(msg: str) -> None:
        pass

@zope.interface.implementer(ISuperSomething)
class SuperSomething(object):
    def hello(self, x: int, y: str) -> None:
        print(f"X: {x}, Y: {y}")

    def greet(self, msg) -> None:
        print(f"Greetings, {msg}")

@zope.interface.implementer(ISomethingSomething)
class SomethingSomething(SuperSomething):
    def wink(self, msg) -> None:
        print(f"{msg} <wink>")

def run(smth: ISomething):
    smth.hello(1, "test")


def greet(smth: ISuperSomething):
    smth.hello(2, "test 2")
    smth.greet("world")


def main() -> None:
    smth = SuperSomething()
    run(smth)

    smthsmth = SomethingSomething()
    smthsmth.greet("bob")
    smthsmth.wink("susanne")
    smthsmth.hello(1, "dude")

    run(smthsmth)


if __name__ == '__main__':
    main()

"""
<output>
</output>
"""



from collections import Iterator
from contextlib import contextmanager
from typing import TypeVar, Generic

_T = TypeVar("_T")


class A(Generic[_T]):
    pass

@contextmanager
def m(x: _T) -> Iterator[A[_T]]:
    ...


with m(7) as x:
    reveal_type(x)
    print(x)


"""
<output>
contextmanager.py:17: note: Revealed type is '__main__.A*[builtins.int*]'
</output>
"""



"""Method parameters is of parameterized type
"""

from typing import List

import zope.interface

class ISomething(zope.interface.Interface):
    def hello(names: List[str]) -> None:
        pass


@zope.interface.implementer(ISomething)
class Something(object):
    def hello(self, names: List[int]) -> None:  # Wrong kind of list
        print(f"Hello, {names}")


"""
<output>
parameterized_types.py:15: error: Argument 1 of "Something" is incompatible with "hello" of supertype "ISomething"; supertype defines the argument type as "List[str]"
parameterized_types.py:15: note: This violates the Liskov substitution principle
parameterized_types.py:15: note: See https://mypy.readthedocs.io/en/stable/common_issues.html#incompatible-overrides
</output>
"""



import zope.interface
import zope.schema


class IBookmark(zope.interface.Interface):
    flag = zope.schema.Bool(required=False)

@zope.interface.implementer(IBookmark)
class Bookmark(object):
    flag = True

def main() -> None:
    bm: IBookmark = Bookmark()
    bm.flag = 343  # Error, expected to be boolean
    bm.flag = None
    bm.flag = True

if __name__ == '__main__':
    main()

"""
<output>
schema_bool.py:14: error: Incompatible types in assignment (expression has type "int", variable has type "Optional[bool]")
</output>
"""



import zope.interface
import zope.schema


class IBookmark(zope.interface.Interface):
    textline = zope.schema.TextLine(
        title='Title',
        description='Optional text line',
        required=False)
    reqtextline = zope.schema.TextLine(
        title='Title',
        description='Implicitly required text line',)
    expreqtextline = zope.schema.TextLine(
        title='Title',
        description='Explicitly required text line',
        required=True)
    asciiline = zope.schema.ASCIILine(
        title='Title',
        description='ASCII-only text line',
        required=False)


@zope.interface.implementer(IBookmark)
class Bookmark(object):
    textline = None
    reqtextline = None
    expreqtextline = None
    asciiline = None


def main() -> None:
    bm: IBookmark = Bookmark()
    bm.textline = 343  # Error, expected to be string
    bm.reqtextline = None  # Error, it is required
    bm.expreqtextline = None  # Error, it is required
    bm.asciiline = True  # Error, it is a string

if __name__ == '__main__':
    main()

"""
<output>
schema_text.py:33: error: Incompatible types in assignment (expression has type "int", variable has type "Optional[str]")
schema_text.py:34: error: Incompatible types in assignment (expression has type "None", variable has type "str")
schema_text.py:35: error: Incompatible types in assignment (expression has type "None", variable has type "str")
schema_text.py:36: error: Incompatible types in assignment (expression has type "bool", variable has type "Optional[str]")
</output>
"""



"""When interface is imported from unknown package, but is inherited by other,
we cannot recognize it as an interface. So standard mypy rules apply, with
all the extra error messages.
"""

import zope.interface

from unknown.interfaces import IUnknownInterface

class IKnownInterface(IUnknownInterface):
    def hello():
        pass

@zope.interface.implementer(IKnownInterface)
class Bookmark(object):
    pass


"""
<output>
interface_unknown_inherit.py:8: error: Cannot find implementation or library stub for module named 'unknown.interfaces'
interface_unknown_inherit.py:8: note: See https://mypy.readthedocs.io/en/latest/running_mypy.html#missing-imports
interface_unknown_inherit.py:11: error: Method must have at least one argument
interface_unknown_inherit.py:14: error: zope.interface.implementer accepts interface, not __main__.IKnownInterface.
interface_unknown_inherit.py:14: error: Make sure you have stubs for all packages that provide interfaces for __main__.IKnownInterface class hierarchy.
</output>
"""



# Stubs for zope.interface.common.sequence (Python 3.6)
#
# NOTE: This dynamically typed stub was automatically generated by stubgen.

from typing import Any, Optional
from zope.interface import Interface

__docformat__: str

class IMinimalSequence(Interface):
    def __getitem__(index: Any) -> None: ...

class IFiniteSequence(IMinimalSequence):
    def __len__() -> None: ...

class IReadSequence(IFiniteSequence):
    def __contains__(item: Any) -> None: ...
    def __lt__(other: Any) -> None: ...
    def __le__(other: Any) -> None: ...
    def __eq__(other: Any) -> None: ...
    def __ne__(other: Any) -> None: ...
    def __gt__(other: Any) -> None: ...
    def __ge__(other: Any) -> None: ...
    def __add__(other: Any) -> None: ...
    def __mul__(n: Any) -> None: ...
    def __rmul__(n: Any) -> None: ...
    def __getslice__(i: Any, j: Any) -> None: ...

class IExtendedReadSequence(IReadSequence):
    def count(item: Any) -> None: ...
    def index(item: Any, *args: Any) -> None: ...

class IUniqueMemberWriteSequence(Interface):
    def __setitem__(index: Any, item: Any) -> None: ...
    def __delitem__(index: Any) -> None: ...
    def __setslice__(i: Any, j: Any, other: Any) -> None: ...
    def __delslice__(i: Any, j: Any) -> None: ...
    def __iadd__(y: Any) -> None: ...
    def append(item: Any) -> None: ...
    def insert(index: Any, item: Any) -> None: ...
    def pop(index: int = ...) -> None: ...
    def remove(item: Any) -> None: ...
    def reverse() -> None: ...
    def sort(cmpfunc: Optional[Any] = ...) -> None: ...
    def extend(iterable: Any) -> None: ...

class IWriteSequence(IUniqueMemberWriteSequence):
    def __imul__(n: Any) -> None: ...

class ISequence(IReadSequence, IWriteSequence): ...
